/*
 *
 * This file is a part of gcd - gcd is an ncurses based audio CD player.
 *
 * Copyright (C) 1998  Geza Herman
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Geza Herman
 * <hg211@ural2.hszk.bme.hu>
 *
 */

#include "gcd.h"
#include "error.h"
#include "display.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/cdrom.h>
#include <sys/ioctl.h>
#include <malloc.h>
#include <stdlib.h> 
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <time.h>

#define  TI_CC 0
#define  TI_RC 1
#define  TI_CR 2
#define  TI_RR 3
#define  TI_NEXT( t ) ( ( t + 1 ) % 4 )

#define	 CR_CDTITLE 0
#define  CR_TITLES  1 
#define  CR_EXT	    2
#define  CR_PROGS   3
#define  CR_NEXT( c ) ( ( c + 1 ) % 4 )
#define  CR_PREV( c ) ( ( c + 3 ) % 4 )

#define  EX_NEXT    1
#define  EX_PREV    2
#define  EX_JUMPEXT 3
#define  EX_QUIT    4

#define  MSF2A( m, s, f ) ( ( m ) * 4500 + ( s ) * 75 + ( f ) )

int	cd_fd;
int	cd_inserted = 0;
int	track_nr = 0;
struct cdrom_volctrl vol;
struct cdrom_subchnl sub;
int	db_support = TRUE;

struct {
    int	    min, sec, frame;
    int     lmin, lsec, lframe;
    u_char  type;
    char   *title;
    char   *ext;            
    int	    played;
} toc[ 100 ];
char   *cdtitle;
char   *cdext; 	
char	playorder[ MAX_PLAYORDER ];
int	tmin, tsec, tframe;
int	f_a, l_a;

int	orderlen = 0;
int	lets_play;
int	first_play;
int	updated, rev_inc;
int	revision;

int	p_status;
int	p_track;
int	time_mode = TI_CC;		     
int	playmode = PM_NORMAL;
int	rep = FALSE;
int	n_ext;

unsigned discid;

int	f_tr, l_tr;

int	t_first, t_curr, t_x;
int	cursor;

int	p_first, p_curr, p_x;
int	pp_track;

void	redraw_all();

int     open_cdrom( char *dev )
{
	return( ( cd_fd = open( dev, O_RDONLY | O_NONBLOCK ) ) != -1 );
}
        
void    close_cdrom()
{
        close( cd_fd );
}

int	handle_error( int e, char *p )
{
	char	buf[ 256 ];
	
	if( e == -1 ) {
	    sprintf( buf, "(%s) %s", p, strerror( errno ) );
	    print_errormsg( buf );
	    refresh();
	    sleep( ERROR_WAIT );
	} else {
	    print_errormsg( NULL );
	}
	return( e );
}

void	update_volume()
{
	handle_error( ioctl( cd_fd, CDROMVOLREAD, &vol ), "READVOL" );
	print_volume( vol.channel0, vol.channel1 );	
}

void	set_volume()
{
	handle_error( ioctl( cd_fd, CDROMVOLCTRL, &vol ), "VOLCTRL" );
}

void	update_playmode()
{
	print_playmode( cd_inserted ? playmode : -1, rep );
}

void	read_entry( int e, int t )
{
	struct cdrom_tocentry	entry;
	
	entry.cdte_track = e;
	entry.cdte_format = CDROM_MSF;
	handle_error( ioctl( cd_fd, CDROMREADTOCENTRY, &entry ), "TOCENTRY" );
	toc[ t ].min = entry.cdte_addr.msf.minute;
	toc[ t ].sec = entry.cdte_addr.msf.second;
	toc[ t ].frame = entry.cdte_addr.msf.frame;
	toc[ t ].type = entry.cdte_ctrl == CDROM_DATA_TRACK ? TR_DATA : TR_AUDIO;
}

void	sub_time( int min1, int sec1, int frame1, int min2, int sec2, int frame2, int *min, int *sec, int *frame )
{
	int	sum = MSF2A( min1 - min2, sec1 - sec2, frame1 - frame2 );
	
	*min = sum / 4500;
	sum %= 4500;
	*sec = sum / 75;
	*frame = sum % 75;
}

void	count_entry( int e )
{
	sub_time( toc[ e + 1 ].min, toc[ e + 1 ].sec, toc[ e + 1 ].frame,
	          toc[ e ].min, toc[ e ].sec, toc[ e ].frame, 
	          &toc[ e ].lmin, &toc[ e ].lsec, &toc[ e ].lframe );
}

int	cddb_sum( int n )
{
	char	buf[ 12 ], *p;
	int	ret = 0;
	
	sprintf( buf, "%u", n );
	for( p = buf; *p != '\0'; p++ ) ret += *p - '0';
	return( ret );
}

unsigned cddb_discid()
{
	int i, t = 0, n = 0;
	
	for( i = 0; i < track_nr; i++ ) n += cddb_sum( toc[ i ].min * 60 + toc[ i ].sec );
	t = ( toc[ track_nr ].min - toc[ 0 ].min ) * 60 + toc[ track_nr ].sec - toc[ 0 ].sec;
	return( ( n & 0xff ) << 24 | t << 8 | track_nr );		
}

void	update_discid()
{
	print_discID( cd_inserted ? discid : 0 );
}

void	read_toc()
{
	struct cdrom_tochdr	hdr;
	int			i;
	
	cd_inserted = ioctl( cd_fd, CDROMREADTOCHDR, &hdr ) != -1;
	if( cd_inserted ) {
	    track_nr = hdr.cdth_trk1;
	    for( i = 1; i <= track_nr; i++ ) read_entry( i, i - 1 );
	    read_entry( CDROM_LEADOUT, track_nr );
	    for( i = 0; i < track_nr; i++ ) count_entry( i );
	    discid = cddb_discid();
	    sub_time( toc[ track_nr ].min, toc[ track_nr ].sec, toc[ track_nr ].frame,
	              toc[ 0 ].min, toc[ 0 ].sec, toc[ 0 ].frame, &tmin, &tsec, &tframe );
	    f_a = MSF2A( toc[ 0 ].min, toc[ 0 ].sec, toc[ 0 ].frame );
	    l_a = MSF2A( toc[ track_nr ].min, toc[ track_nr ].sec, toc[ track_nr ].frame );	
	} else {
	    track_nr = 0;
	}
}

void	clear_toc()
{
	int	i;
	
	for( i = 0; i < 100; i++ ) {
	    toc[ i ].title = toc[ i ].ext = NULL;
	    toc[ i ].played = FALSE;
	}
	cdtitle = NULL;
	cdext = NULL;
	orderlen = 0;
	t_first = t_curr = t_x = 0;
	pp_track = p_first = p_curr = p_x = 0;
	cursor = CR_TITLES;
	updated = FALSE;
	rev_inc = FALSE;
	n_ext = 0;
	lets_play = first_play = FALSE;
	revision = -1;
}

void	free_toc()
{
	int	i;
		
	for( i = 0; i < 100; i++ ) {
	     if( toc[ i ].title ) free( toc[ i ].title );
	     if( toc[ i ].ext ) free( toc[ i ].ext );
	}
	if( cdtitle ) free( cdtitle );
	if( cdext ) free( cdext );
}

void	read_orderline( char *buf )
{
	char   *c;
	
	while( *buf ) {
	    playorder[ orderlen++ ] = atoi( buf );
	    c = strchr( buf, ',' );
	    buf = c == NULL ? buf + strlen( buf ) : c + 1;
	}
}

int	read_cddb( int homedir )
{
	char	buf[ 256 ];
	char    ext[ MAX_EXT + 1 ];
	FILE   *f;
	int	n;
	char   *c;	
	int	lastext = -2;
	char   *cext = ext;	
	
	if( homedir ) {
	    sprintf( buf, "%s/.gcd/%08x", getenv( "HOME" ), discid );
	} else {
	    sprintf( buf, CDDB_DIR"/%08x", discid );
	}        

	f = fopen( buf, "rt" );
	if( f != NULL ) {
	    while( !feof( f ) ) {
		buf[ 0 ] = '\0';
	        fgets( buf, 256, f );
	        if( ( c = strchr( buf, '\n' ) ) != NULL ) *c = '\0';
	        if( *buf != '#' ) {
	            if( strncmp( buf, "DTITLE", 6 ) == 0 ) {
	                if( buf[ 8 ] != '\0' ) {
	                    cdtitle = malloc( strlen( buf + 7 ) + 1 );
	                    strcpy( cdtitle, buf + 7 );
	                }
	            }
	            if( strncmp( buf, "TTITLE", 6 ) == 0 ) {
	                if( buf[ 8 ] != '\0' ) {
	                    n = atoi( buf + 6 );
	                    c = strchr( buf, '=' ) + 1;
	                    toc[ n ].title = malloc( strlen( c ) + 1 );
	                    strcpy( toc[ n ].title, c );
	                }
	            }
	            if( strncmp( buf, "EXTD", 4 ) == 0 ) {
	                memcpy( cext, buf + 5, strlen( buf + 5 ) );
	                cext += strlen( buf + 5 );
	                lastext = -1;
	            }
	            if( strncmp( buf, "EXTT", 4 ) == 0 ) {
	                if( lastext == -1 ) {
	                    if( cext > ext ) {
	                        cdext = malloc( cext - ext + 1 );
	                        memcpy( cdext, ext, cext - ext );
	                        cdext[ cext - ext ] = '\0';
	                        cext = ext;
	                    }
	                    lastext = 0;
	                }
	                n = atoi( buf + 4 );
	                c = strchr( buf, '=' ) + 1;
	                if( lastext != n ) {
	                    if( cext > ext ) {
	                        toc[ lastext ].ext = malloc( cext - ext + 1 );
	                        memcpy( toc[ lastext ].ext, ext, cext - ext );
	                        toc[ lastext ].ext[ cext - ext ] = '\0';
	                        cext = ext;
	                    }    
	                    lastext = n;
	                }
                        memcpy( cext, c, strlen( c ) );
                        cext += strlen( c );
	            }
	            if( strncmp( buf, "PLAYORDER", 9 ) == 0 ) {
	                if( cext > ext ) {
	                    toc[ lastext ].ext = malloc( cext - ext + 1 );
	                    memcpy( toc[ lastext ].ext, ext, cext - ext );
	                    toc[ lastext ].ext[ cext - ext ] = '\0';
	                    cext = ext;
	                }
	                read_orderline( buf + 10 );
	            }        
	        } else {
		    if( strncmp( buf, "# Revision: ", 12 ) == 0 ) {
		        revision = atoi( buf + 12 );
		    }    
	        }                           
	    }
	    fclose( f );
	    return( TRUE );
	} else {
	    if( errno != ENOENT && db_support ) {
	        handle_error( -1, "READCDDB" );
	    }
	    return( FALSE );
	}        
}

void	write_cddb_ext( FILE *f, char *s, char *prefix )
{
	int	l, pl = strlen( prefix );
	
	if( s != NULL ) {
	    while( *s ) {
	        l = pl + 1;
	        fprintf( f, "%s=", prefix );
	        while( l < 79 && *s ) {
	            fputc( *s++, f );
	            l++;
	        }    
	        fputc( '\n', f );
	    }
	} else {
	    fprintf( f, "%s=\n", prefix );
	}              
}

void	write_cddb_playorder( FILE *f )
{
	int	i;
	char	buf[ 8 ];
	int	l = 10, sl, first = TRUE;
	
	fprintf( f, "PLAYORDER=" );
	if( orderlen == 0 ) {
	    fprintf( f, "\n" );
	} else {
	    for( i = 0; i < orderlen; i++ ) {
	        sprintf( buf, "%d", playorder[ i ] );
	        sl = strlen( buf );
	        l += sl;
	        if( !first ) l++;
	        if( l >= 80 ) {
	            first = TRUE;
	            fprintf( f, "\nPLAYORDER=" );
	            l = 10 + sl;
	        }    
	        if( first ) {
	            first = FALSE;
	            fprintf( f, "%s", buf );
	        } else {
	            fprintf( f, ",%s", buf );
	        }
	    }
	}    
}

void	write_cddb( int nc_act )
{
	char	buf[ 256 ];
	FILE   *f;
	int	i; 
	
	sprintf( buf, "%s/.gcd/%08x", getenv( "HOME" ), discid );
	f = fopen( buf, "wt" );
	if( f != NULL ) {
	    fprintf( f, "# xmcd, keep smiling!\n" );
	    fprintf( f, "#\n" );
	    fprintf( f, "# Track frame offsets:\n" );
	    for( i = 0; i < track_nr; i++ ) fprintf( f, "#       %d\n", MSF2A( toc[ i ].min, toc[ i ].sec, toc[ i ].frame ) );
	    fprintf( f, "#\n" );
	    fprintf( f, "# Disc length: %d seconds\n", tmin * 60 + tsec );
	    fprintf( f, "#\n" );
	    if( rev_inc ) revision++;
	    fprintf( f, "# Revision: %d\n", revision );
	    fprintf( f, "# Submitted via: gcd "VERSION", Geza\'s CD player\n" );
	    fprintf( f, "#\n" );
	    
	    fprintf( f, "DISCID=%08x\n", discid );
	    fprintf( f, "DTITLE=" );
	    if( cdtitle ) fprintf( f, "%s", cdtitle ); 
	    fprintf( f, "\n" );
	    for( i = 0; i < track_nr; i++ ) {
	        fprintf( f, "TTITLE%d=", i );
	        if( toc[ i ].title ) fprintf( f, "%s", toc[ i ].title );
	        fprintf( f, "\n" );
	    }
	    write_cddb_ext( f, cdext, "EXTD" );
	    for( i = 0; i < track_nr; i++ ) {
	        sprintf( buf, "EXTT%d", i );
	        write_cddb_ext( f, toc[ i ].ext, buf );
	    }
	    write_cddb_playorder( f );
	    fclose( f );
	    if( !nc_act ) fprintf( stderr, " Successful.\n" );
	} else {
	    if( nc_act ) {
	        handle_error( -1, "WRITECDDB" );
	    } else {
	        fprintf( stderr, " Failed.\n" );
	        perror( "WRITECDDB" );
	    }         
	}    
}

void	update_cdtotal()
{
	if( cd_inserted ) {
	    print_cdtotal( tmin, tsec, tframe );
	} else {
	    print_cdtotal( -1, 0, 0 );
	}         
}

void	update_toc()
{
	read_toc();
	free_toc();
	clear_toc();
	if( cd_inserted ) if( !read_cddb( TRUE ) ) read_cddb( FALSE );
}

void	count_time( int c, int min, int sec, int frame, int lmin, int lsec, int lframe, int *rmin, int *rsec, int *rframe )
{
	if( c ) {
	    sub_time( lmin, lsec, lframe, min, sec, frame, rmin, rsec, rframe );
	} else {    
	    *rmin = min; *rsec = sec; *rframe = frame;
	}    
}

void	update_subchnl()
{
	int		     min, sec, frame;
	int		     minus;	
	
	if( cd_inserted ) {
            sub.cdsc_format = CDROM_MSF;
	    handle_error( ioctl( cd_fd, CDROMSUBCHNL, &sub ), "SUBCHNL" );
	    switch( sub.cdsc_audiostatus ) {
	        case CDROM_AUDIO_PLAY:      p_status = ST_PLAYING; break;
	        case CDROM_AUDIO_PAUSED:    p_status = ST_PAUSED;  break;
	        case CDROM_AUDIO_NO_STATUS:
	        case CDROM_AUDIO_ERROR:
	        case CDROM_AUDIO_COMPLETED: p_status = ST_STOPPED; break;
	        default:                    p_status = ST_UNKNOWN;
	    }
	    p_track = sub.cdsc_trk - 1;
	    minus = time_mode == TI_RC || time_mode == TI_RR;
            count_time( minus,
	                sub.cdsc_absaddr.msf.minute, sub.cdsc_absaddr.msf.second, sub.cdsc_absaddr.msf.frame,
		        toc[ track_nr ].min, toc[ track_nr ].sec, toc[ track_nr ].frame,
		        &min, &sec, &frame );
            print_cdtime( min, sec, minus ? '-' : ' ' );
            minus = time_mode == TI_CR || time_mode == TI_RR; 
            count_time( minus,
	                sub.cdsc_reladdr.msf.minute, sub.cdsc_reladdr.msf.second, sub.cdsc_reladdr.msf.frame,
	                toc[ p_track ].lmin, toc[ p_track ].lsec, toc[ p_track ].lframe,
	                &min, &sec, &frame );
	    print_tracktime( min, sec, minus ? '-' : ' ' );
	} else {
	    p_status = ST_NODISK;
	    print_cdtime( -1, 0, ' ' );
	    print_tracktime( -1, 0, ' ' );
	}        
		    	
	print_state( p_status );	
}

void	open_tray()
{
	if( handle_error( ioctl( cd_fd, CDROMEJECT ), "OPENTRAY" ) != -1 ) {
	    if( updated && db_support ) write_cddb( TRUE );
	    update_toc();
	    redraw_all();
	}    
}

void	close_tray()
{
	if( handle_error( ioctl( cd_fd, CDROMCLOSETRAY ), "CLOSETRAY" ) != -1 ) {
	    update_toc();
	    redraw_all();
	}    	
}

void	print_titles( int must_update, int ignore )
{
	int	h = track_nr - t_first > title_maxy() ? title_maxy() : track_nr - t_first;
	int	i;
	int	t = t_first;
	
	static int	o_playmode = -1;
	static int	o_t_curr   = -1;
	static int	o_t_x      = -1;
	static int	o_cursor   = -1;
	static int	o_p_status = -1;
	static int	o_p_track  = -1;
	static int	o_ignore   = -1;
	
	if( !must_update && o_playmode == playmode && o_t_curr == t_curr && o_t_x == t_x && 
	    o_cursor == cursor && o_p_status == p_status && o_p_track == p_track && o_ignore == ignore ) return;
	o_playmode = playmode; o_t_curr = t_curr;
	o_t_x = t_x; o_cursor = cursor; o_p_status = p_status;
	o_p_track = p_track; o_ignore = ignore;

	for( i = 0; i < h; i++, t++ ) {
	    print_title( i, t, toc[ t ].type == TR_AUDIO && ( !toc[ t ].played || playmode != PM_SHUFFLE ), 
	                 toc[ t ].title + ( t == t_curr ? t_x : 0 ), 
	                 cursor == CR_TITLES ? t_curr : -1, 
	                 ( p_status == ST_PLAYING || p_status == ST_PAUSED ) ? p_track : -1,
	                 toc[ t ].min, toc[ t ].sec, toc[ t ].frame,
	                 toc[ t ].lmin, toc[ t ].lsec, toc[ t ].lframe, i == ignore );
	}      
	for( ; i < title_maxy(); i++ ) delete_title( i );
}

void	print_progs( int must_update )
{
	int	h = orderlen - p_first > prog_maxy() ? prog_maxy() : orderlen - p_first;
	int	i;
	int	t = p_first;
	char	buf[ 12 ];
	char   *c;
	
	static int	o_p_curr   = -1;
	static int	o_p_x      = -1; 
	static int	o_cursor   = -1; 
	static int	o_p_status = -1;
	static int	o_playmode = -1;
	static int	o_pp_track = -1;
	
	if( !must_update && o_p_curr == p_curr && o_p_x == p_x && o_cursor == cursor &&
	    o_p_status == p_status && o_playmode == playmode && o_pp_track == pp_track ) return;
	o_p_curr = p_curr; o_p_x = p_x; o_cursor = cursor;
	o_p_status = p_status; o_playmode = playmode; o_pp_track = pp_track;
	
	for( i = 0; i < h; i++, t++ ) {
	    if( ( c = toc[ ( int )playorder[ t ] ].title + ( t == p_curr ? p_x : 0 ) ) == NULL ) {
	        sprintf( buf, "-Track %02d.-", playorder[ t ] + 1 );
	        c = buf;       
	    }    
	    print_prog( i, t, c, 
	                cursor == CR_PROGS ? p_curr : -1, 
	                ( p_status == ST_PLAYING || p_status == ST_PAUSED ) && playmode == PM_PROGRAM ? pp_track - 1 : -1 );
	}
	for( ; i < prog_maxy(); i++ ) delete_prog( i );
}

int	last_playable( int f )
{
	while( f < track_nr && toc[ f ].type != TR_DATA ) f++;
	return( f - 1 );
}

void	cdrom_play( struct cdrom_msf *msf )
{
	handle_error( ioctl( cd_fd, CDROMPLAYMSF, msf ), "PLAY" );
}

void	cdrom_stop()
{
	handle_error( ioctl( cd_fd, CDROMSTOP ), "STOP" );
}

void	cdrom_pause()
{
	handle_error( ioctl( cd_fd, CDROMPAUSE ), "PAUSE" );
}

void	cdrom_resume()
{
	handle_error( ioctl( cd_fd, CDROMRESUME ), "RESUME" );
}

int	next_shuffle()
{
	int	found = FALSE;
	int	i = 0;
	
	while( !found && i < track_nr ) {
	    if( !toc[ i ].played && toc[ i ].type == TR_AUDIO ) {
	        found = TRUE;
	    } else {
	        i++;
	    }
	}
	if( found ) {
	    do {
	        i = ( rand() >> 8 ) % track_nr;
	    } while( toc[ i ].played || toc[ i ].type != TR_AUDIO );
	    return( i );
	} else {
	    return( -1 );
	}                
}

int	next_audio( int tr ) 
{
	int	found = FALSE;

	while( tr < track_nr && !found ) {
	    if( toc[ tr ].type == TR_AUDIO ) {
	        found = TRUE;
	    } else {
	        tr++;
	    }
	}
	return( found ? tr : -1 );            
}

void	fill_msf( int b, int e, struct cdrom_msf *msf )
{
	msf->cdmsf_min0   = toc[ b ].min;
	msf->cdmsf_sec0   = toc[ b ].sec;
	msf->cdmsf_frame0 = toc[ b ].frame;
	msf->cdmsf_min1   = toc[ e + 1 ].min;
	msf->cdmsf_sec1   = toc[ e + 1 ].sec;
	msf->cdmsf_frame1 = toc[ e + 1 ].frame;
}

int	msfchk( struct cdrom_msf *msf )
{
	int	a0 = MSF2A( msf->cdmsf_min0, msf->cdmsf_sec0, msf->cdmsf_frame0 );
	int	a1 = MSF2A( msf->cdmsf_min1, msf->cdmsf_sec1, msf->cdmsf_frame1 );
	
	if( a0 < f_a || a0 > l_a ) return( FALSE );
	if( a1 < f_a || a1 > l_a ) return( FALSE );
	return( a0 < a1 );
}

void	fill_msfi( int t, struct cdrom_msf *msf )
{
	int	min, sec, frame;
	
	msf->cdmsf_min0   = toc[ t ].min;
	msf->cdmsf_sec0   = toc[ t ].sec;
	msf->cdmsf_frame0 = toc[ t ].frame;
	sub_time( toc[ t ].min, toc[ t ].sec, toc[ t ].frame, 0, -INTRO_TIME, 0, &min, &sec, &frame );
	if( min > toc[ t + 1 ].min ) min = toc[ t + 1 ].min;
	if( min == toc[ t + 1 ].min && sec > toc[ t + 1 ].sec ) sec = toc[ t + 1 ].sec;
	if( min == toc[ t + 1 ].min && sec == toc[ t + 1 ].sec && frame > toc[ t + 1 ].frame ) frame = toc[ t + 1 ].frame;
	msf->cdmsf_min1   = min;
	msf->cdmsf_sec1   = sec;
	msf->cdmsf_frame1 = frame;
}

void	do_play()
{
	struct cdrom_msf	msf;

//	if( ( playmode == PM_NORMAL && !first_play && !rep ) || playmode == PM_PROGRAM ) return;
	switch( playmode ) {
	    case PM_NORMAL:
                if( first_play || rep ) {
                    fill_msf( f_tr, l_tr, &msf );
                    cdrom_stop();
                    cdrom_play( &msf );
                } else {
                    lets_play = FALSE;
                }       
                break;
            case PM_PROGRAM:
                if( rep && pp_track == orderlen ) pp_track = 0;
                if( pp_track < orderlen ) {
                    fill_msf( playorder[ pp_track ], playorder[ pp_track ], &msf );
                    cdrom_stop();
                    cdrom_play( &msf );
                    pp_track++;
                } else {
                    lets_play = FALSE;
                }        
                break;
            case PM_SHUFFLE: {
                int	tr = next_shuffle();
                int	i;
                
                if( tr == -1 ) {
                    if( rep ) {
                        for( i = 0; i < track_nr; i++ ) toc[ i ].played = FALSE;
                        tr = next_shuffle();
                        if( tr == -1 ) lets_play = FALSE;
                    } else {
                        lets_play = FALSE;
                    }
                }
		if( lets_play ) {
                    toc[ tr ].played = TRUE;
                    fill_msf( tr, tr, &msf );
                    cdrom_stop();            
		    cdrom_play( &msf );
		}    		                    
                break;
            }
            case PM_INTRO: {
		pp_track = next_audio( pp_track );
		if( pp_track == -1 ) {
		    if( rep ) {
		        pp_track = next_audio( 0 );
		        if( pp_track == -1 ) lets_play = FALSE;
		    } else {
		        lets_play = FALSE;
		    }
		}
		if( lets_play ) {
		    fill_msfi( pp_track, &msf );
		    cdrom_stop();
		    cdrom_play( &msf );
		    pp_track++;
		}
		break;
	    }	            
        } 
	first_play = FALSE;
}

void	inc_volume( u_char *vol )
{
	int	v = *vol;
	
	v = ( v * 1.2 ) + 1;
	if( v > 255 ) v = 255;
	*vol = v;
}

void	dec_volume( u_char *vol )
{
	*vol *= 0.83333;
}

void	skip( int amount )
{
	struct cdrom_msf	msf;
	int			m, s, f;

	if( cd_inserted && lets_play ) {
	    sub_time( sub.cdsc_absaddr.msf.minute, sub.cdsc_absaddr.msf.second, sub.cdsc_absaddr.msf.frame,
	              0, -amount, 0, &m, &s, &f );
	    cdrom_stop();
	    if( m < 0 || s < 0 || f < 0 ) return;
	    msf.cdmsf_min0   = m;
	    msf.cdmsf_sec0   = s;
	    msf.cdmsf_frame0 = f;
	    switch( playmode ) {
	        case PM_NORMAL:
	            msf.cdmsf_min1   = toc[ l_tr + 1 ].min;
	            msf.cdmsf_sec1   = toc[ l_tr + 1 ].sec;
	            msf.cdmsf_frame1 = toc[ l_tr + 1 ].frame;
	            break;
	        case PM_PROGRAM:
	            msf.cdmsf_min1   = toc[ ( int )playorder[ pp_track - 1 ] + 1 ].min;    
	            msf.cdmsf_sec1   = toc[ ( int )playorder[ pp_track - 1 ] + 1 ].sec;    
	            msf.cdmsf_frame1 = toc[ ( int )playorder[ pp_track - 1 ] + 1 ].frame;
	            break;
	        case PM_SHUFFLE:
	            msf.cdmsf_min1   = toc[ p_track + 1 ].min;
	            msf.cdmsf_sec1   = toc[ p_track + 1 ].sec;
	            msf.cdmsf_frame1 = toc[ p_track + 1 ].frame;
	            break;        
	    }        
	    if( msfchk( &msf ) ) cdrom_play( &msf );
	}
}

int	handle_gen_keys( int ch )
{
	int	quit = FALSE;

	switch( ch ) {
            case 'm':
	    case 'M':
	        if( !lets_play && cd_inserted ) {
	            playmode = PM_NEXT( playmode );
	            update_playmode();
	        }    
	        break;
	    case 'r':
	    case 'R':
	        if( cd_inserted ) {
	            rep = !rep;
	            print_playmode( playmode, rep );
	        }    
	        break;    
	    case 'e':
	    case 'E': open_tray(); break;
	    case 'c':
	    case 'C': close_tray(); break;
	    case 's':
	    case 'S': 
	        if( cd_inserted ) {
	            cdrom_stop(); 
	            lets_play = FALSE;
	        }    
	        break;
	    case ' ':
	        if( cd_inserted ) {
	            if( p_status == ST_PAUSED ) { 
	                cdrom_resume();
	            } else {
	                if( p_status == ST_PLAYING ) {
	                    cdrom_pause();
	                }
	            }
	        }
	        break;                    
	    case '.': skip( B_SKIP ); break;
	    case ',': skip( -B_SKIP ); break;
	    case '>': skip( S_SKIP ); break;
	    case '<': skip( -S_SKIP ); break;    
	    case 't':
	    case 'T':
	        if( cd_inserted ) time_mode = TI_NEXT( time_mode ); 
	        break;
	    case 'x':
	    case 'X':
	        if( cd_inserted && db_support ) quit = EX_JUMPEXT;
	        break;       
            case MKEY_ENTER:
	        if( !cd_inserted ) break;
	        switch( playmode ) {
	            case PM_NORMAL:
	                if( toc[ t_curr ].type == TR_AUDIO ) {
	                    f_tr = t_curr;
	                    l_tr = last_playable( t_curr );
	                    lets_play = first_play = TRUE;
	                }
	                break;
	            case PM_PROGRAM:
	                if( orderlen > 0 ) {
	                    pp_track = cursor == CR_PROGS ? p_curr : 0;
	                    lets_play = first_play = TRUE;
	                }    
	                break;
	            case PM_SHUFFLE: {
	                int	i;
	                
	                for( i = 0; i < track_nr; i++ ) toc[ i ].played = FALSE;
	                lets_play = first_play = TRUE;
	                break;
	            }
	            case PM_INTRO:
	                pp_track = cursor == CR_TITLES ? t_curr : 0;
	                lets_play = first_play = TRUE;
	                break;    
	        } 
	        break;
	    case '\\' : inc_volume( &vol.channel0 ); set_volume(); break;
	    case '=': dec_volume( &vol.channel0 ); set_volume(); break;
	    case ']' : inc_volume( &vol.channel0 ); inc_volume( &vol.channel1 ); set_volume(); break;
	    case '[' : dec_volume( &vol.channel0 ); dec_volume( &vol.channel1 ); set_volume(); break;
	    case '\'': inc_volume( &vol.channel1 ); set_volume(); break;
	    case ';' : dec_volume( &vol.channel1 ); set_volume(); break;
	    case MKEY_TAB:
	        if( cd_inserted ) quit = EX_NEXT; 
	        break;    
            case MKEY_ESC: quit = EX_QUIT; break;
	}
	return( quit );
}

void	handle_play()
{
	if( lets_play && ( ( p_status != ST_PLAYING && p_status != ST_PAUSED ) || first_play ) ) do_play();
}

void	input_title()
{
	char	title[ MAX_TITLE + 1 ];
	int	l;
	int	in_ex;
	struct input_inf i;
	
	if( toc[ t_curr ].title ) {
	    strncpy( title, toc[ t_curr ].title, MAX_TITLE );
	    title[ MAX_TITLE ] = '\0';
	} else {
	    title[ 0 ] = '\0';
	}    

	init_input( &i );

	do {
	    update_subchnl();
	    update_volume();
	    print_titles( FALSE, t_curr );
	    print_progs( FALSE );
	    handle_play();
	    
	    set_color( toc[ t_curr ].type == TR_AUDIO, t_curr, 
	               ( p_status == ST_PLAYING || p_status == ST_PAUSED ) ? p_track : -1, t_curr );
	    in_ex = input( title_y() + t_curr - t_first, title_x(), title_maxx(), title, MAX_TITLE, &i );
	} while( in_ex == IN_AGAIN );
	    
	if( in_ex == IN_ACCEPT && i.modified ) {
	    updated = rev_inc = TRUE;
	    t_x = 0;
	    if( toc[ t_curr ].title ) free( toc[ t_curr ].title );
	    l = strlen( title );
	    if( l != 0 ) {
	        toc[ t_curr ].title = malloc( l + 1 );
	        strcpy( toc[ t_curr ].title, title );
	    } else {
	        toc[ t_curr ].title = NULL;
	    }
	    print_titles( TRUE, -1 );
	    print_progs( TRUE );
	}        
}

void	clear_title()
{
	if( toc[ t_curr ].title ) {
	    free( toc[ t_curr ].title );
	    toc[ t_curr ].title = NULL;
	    print_titles( TRUE, -1 );
	    print_progs( TRUE );
	}    
}

int	handle_ext()
{
	char			ext[ MAX_EXT + 1 ];
	char		      **e = n_ext == -1 ? &cdext : &( toc[ n_ext ].ext );
	struct input_inf	i;
	int			in_ex;
	int			l;
	
	if( !db_support ) return( EX_NEXT );
	
	if( *e != NULL ) {
	    strncpy( ext, *e, MAX_EXT );
	    ext[ MAX_EXT ] = '\0';
	} else {
	    ext[ 0 ] = '\0';
	}
	
	init_input( &i );
	do {
	    update_subchnl();
	    update_volume();
	    print_titles( FALSE, -1 );
	    print_progs( FALSE );
	    handle_play();

	    in_ex = input_ext( ext, MAX_EXT, &i );
	} while( in_ex == IN_AGAIN );
	
	if( in_ex == IN_ACCEPT && i.modified ) {
	    updated = rev_inc = TRUE;
	    if( *e ) free( *e );
	    l = strlen( ext );
	    if( l != 0 ) {
	        *e = malloc( l + 1 );
	        strcpy( *e, ext );
	    } else {
		*e = NULL;
	    }
	    print_ext( 0, 0, *e );
	}
	return( in_ex == IN_ACCEPT ? EX_NEXT : EX_QUIT );                
}

int	handle_progs()
{
	int	quit = FALSE;
	int	old = -1;
	int	ch;
	int	oci = !cd_inserted;
	int	pmust = FALSE;
	
	if( orderlen == 0 ) return( EX_NEXT );
	do {
	    if( old != p_curr || oci != cd_inserted ) {
	        old = p_curr;
	        oci = cd_inserted;
	        if( p_first > p_curr ) p_first = p_curr;
	        if( p_first + prog_maxy() - 1 < p_curr ) p_first = p_curr - prog_maxy() + 1;
	        p_x = 0;
	    }

	    update_subchnl();
	    update_volume();
	    print_titles( FALSE, -1 );
	    print_progs( pmust );
	    pmust = FALSE;

	    switch( ch = getch() ) {
	        case KEY_LEFT: if( p_x > 0 ) p_x--; break;
	        case KEY_RIGHT: if( toc[ ( int )playorder[ p_curr ] ].title && strlen( toc[ ( int )playorder[ p_curr ] ].title ) - p_x > prog_maxx() ) p_x++; break;
	        case KEY_UP: if( p_curr > 0 ) p_curr--; break;
	        case KEY_DOWN: if( p_curr < orderlen - 1 ) p_curr++; break;
		case '-': 
		    if( cd_inserted && orderlen > 0 && ( p_status == ST_STOPPED || ( p_status != ST_STOPPED && playmode != PM_PROGRAM ) ) ) {
		        p_x = 0;
		        if( p_curr == orderlen - 1 ) {
		            if( p_curr > 0 ) p_curr--;
		        } else {     
		            memmove( playorder + p_curr, playorder + p_curr + 1, orderlen - p_curr - 1 );
		        }    
		        orderlen--;
		        pmust = TRUE;
		        updated = TRUE;
		    }
		    break;     
	        default: quit = handle_gen_keys( ch );        
	    }
	    if( orderlen == 0 ) quit = EX_NEXT;
	    handle_play();
	} while( !quit );
	p_x = 0;
	return( quit );        
}

void	input_cdtitle()
{
	char	title[ MAX_TITLE + 1 ];
	int	l;
	int	in_ex;
	struct input_inf i;
	
	if( cdtitle ) {
	    strncpy( title, cdtitle, MAX_TITLE );
	    title[ MAX_TITLE ] = '\0';
	} else {
	    title[ 0 ] = '\0';
	}    

	init_input( &i );

	do {
	    update_subchnl();
	    update_volume();
	    print_titles( FALSE, -1 );
	    print_progs( FALSE );
	    handle_play();
	    
	    set_color( TRUE, 0, 0, 0 );
	    in_ex = input( 0, 11, cdtitle_max(), title, MAX_TITLE, &i );
	} while( in_ex == IN_AGAIN );
	    
	if( in_ex == IN_ACCEPT && i.modified ) {
	    updated = rev_inc = TRUE;
	    if( cdtitle ) free( cdtitle );
	    l = strlen( title );
	    if( l != 0 ) {
	        cdtitle = malloc( l + 1 );
	        strcpy( cdtitle, title );
	    } else {
	        cdtitle = NULL;
	    }
	}        
}

void	clear_cdtitle()
{
	if( cdtitle ) {
	    free( cdtitle );
	    cdtitle = NULL;
	}    
}


int	handle_cdtitle()
{
	int	cd_x = 0;
	int	quit = FALSE;
	int	ch;
	
	print_wh_ext( "CD" );
	print_ext( 0, 0, cdext );
	n_ext = -1;
	do {
	    print_cdtitle( cdtitle + cd_x, TRUE );
	    update_subchnl();
	    update_volume();
	    print_titles( FALSE, -1 );
	    print_progs( FALSE );

	    switch( ch = getch() ) {
	        case KEY_LEFT: if( cd_x > 0 ) cd_x--; break;
	        case KEY_RIGHT: if( cdtitle && strlen( cdtitle ) - cd_x > cdtitle_max() ) cd_x++; break;
		case KEY_IC: if( cd_inserted && db_support ) { input_cdtitle(); cd_x = 0; } break;
		case KEY_DC: if( cd_inserted && db_support ) { clear_cdtitle(); cd_x = 0; } break;
		default: quit = handle_gen_keys( ch );
	    }	
	} while( !quit );
	print_cdtitle( cdtitle, FALSE );
	return( quit );
}

int	handle_titles()
{
	int	quit = FALSE;
	int	old = -1;
	int	ch;
	int	oci = !cd_inserted;
	char	buf[ 12 ];
	int	pmust = FALSE, emust = FALSE;
	
	do {
	    if( old != t_curr || oci != cd_inserted || emust ) {
	        old = t_curr;
	        oci = cd_inserted;
	        if( t_first > t_curr ) t_first = t_curr;
	        if( t_first + title_maxy() - 1 < t_curr ) t_first = t_curr - title_maxy() + 1;
	        t_x = 0;
	        n_ext = t_curr;
	        if( toc[ n_ext ].title == NULL ) {
	            sprintf( buf, "-Track %02d.-", n_ext + 1 );
	            print_wh_ext( buf );
	        } else {    
	            print_wh_ext( toc[ n_ext ].title );
	        }    
	        print_ext( 0, 0, toc[ n_ext ].ext );
	        emust = FALSE;    
	    }
	    
	    update_subchnl();
	    update_volume();
	    print_titles( FALSE, -1 );
	    print_progs( pmust );
	    pmust = FALSE;

	    switch( ch = getch() ) {
	        case KEY_LEFT: if( t_x > 0 ) t_x--; break;
	        case KEY_RIGHT: if( toc[ t_curr ].title && strlen( toc[ t_curr ].title ) - t_x > title_maxx() ) t_x++; break;
	        case KEY_UP: if( t_curr > 0 ) t_curr--; break;
	        case KEY_DOWN: if( t_curr < track_nr - 1 ) t_curr++; break;
	        case KEY_IC: if( cd_inserted && db_support ) { input_title(); emust = TRUE; }; break;
	        case KEY_DC: if( cd_inserted && db_support ) { clear_title(); emust = TRUE; t_x = 0; }; break;
		case '+': 
		     if( cd_inserted && orderlen < MAX_PLAYORDER && toc[ t_curr ].type == TR_AUDIO ) { 
		         playorder[ orderlen++ ] = t_curr; 
		         pmust = TRUE;
		         if( p_first + prog_maxy() < orderlen ) { p_first++; p_curr++; }
		         updated = TRUE; 
		     } 
		     break;
	        default: quit = handle_gen_keys( ch );        
	    }
	    handle_play();
	} while( !quit );
        t_x = 0;
        return( quit );
}

void	init_gcddir()
{
	struct stat	buf;
	char		name[ 224 ];
	char		errormsg[ 256 ];
	
	if( getenv( "HOME" ) == NULL ) {
	    fprintf( stderr, "Hey! $HOME is not set!\n" );
	    exit( ERR_OPNDIR );
	}    
	sprintf( name, "%s/.gcd/", getenv( "HOME" ) );
	if( stat( name, &buf ) == -1 ) {
	    if( errno == ENOENT ) {
	        fprintf( stderr, "Directory %s doesn't exist, trying to create it...", name ); 
                if( mkdir( name, 0777 ) == 0 ) {
                    fprintf( stderr, " Successful.\n" );
                    printf( "%s directory created. Press any key to continue...", name );
                    fflush( stdout );
		    getchar();	
                } else {
                    fprintf( stderr, " Failed\n" );
                    sprintf( errormsg, "Error creating %s", name );
                    perror( errormsg );
                    db_support = FALSE;
                }        
            } else {
                sprintf( errormsg, "Can't open directory %s", name );
                perror( errormsg );
                db_support = FALSE;
            }      
	}
}

void	redraw_all()
{
	clear();
	init_screen();
	update_cdtotal();
	update_discid();
	update_playmode();
	print_cdtitle( cdtitle, cursor == CR_CDTITLE );
	update_subchnl();
	update_volume();
	print_titles( TRUE, -1 );
	print_progs( TRUE );
	refresh();
}

void	sighandler( int s )
{
	done_display();
	if( s != SIGWINCH ) {
	    close_cdrom();
	    fprintf( stderr, "Signal %d caught, exiting...\n", s );
	    if( updated && db_support ) {
	        fprintf( stderr, "Trying to save modified data..." );
		write_cddb( FALSE );
	    }    
	    exit( s + ERR_SIGNAL );
	} else {
	    init_display();
	    redraw_all();
	}    
} 


void	init_sighandler()
{
	int	sig[ 6 ] = { SIGHUP, SIGINT, SIGQUIT, SIGSEGV, SIGTERM, SIGWINCH };
	int	i;
	
	for( i = 0; i < 6; i++ ) signal( sig[ i ], sighandler );
}

void	init( int argc, char *argv[] )
{
	time_t	t;
	char	buf[ 256 ];

	if( !open_cdrom( argc == 1 ? DEFAULT_CDROM : argv[ 1 ] ) ) {
	    sprintf( buf, "Error opening cdrom device(%s)", argc == 1 ? DEFAULT_CDROM : argv[ 1 ] );
	    perror( buf );
	    exit( ERR_OPNDEV );
	}
	time( &t );
	srand( ( unsigned )t );
	
	init_gcddir();
	init_display();
	init_screen();
	init_sighandler();
	clear_toc();
	update_toc();
	update_cdtotal();
	update_discid();
	update_playmode();
	print_cdtitle( cdtitle, FALSE );
	print_titles( TRUE, -1 );
	print_progs( TRUE );
	update_subchnl();
	if( p_status == ST_PLAYING || p_status == ST_PAUSED ) {
	    lets_play = TRUE;
	    f_tr = p_track;
	    l_tr = last_playable( f_tr );
	}    
}                

int 	main( int argc, char *argv[] )
{
	int	quit = FALSE;
	int	exitcode;

	init( argc, argv );
	do {
	    switch( cursor ) {
		case CR_CDTITLE: exitcode = handle_cdtitle(); break;
	        case CR_TITLES: exitcode = handle_titles(); break;
	        case CR_EXT: exitcode = handle_ext(); break;
	        case CR_PROGS: exitcode = handle_progs(); break; 
	    }    
	    switch( exitcode ) {
	        case EX_NEXT: cursor = CR_NEXT( cursor ); break;
	        case EX_JUMPEXT: cursor = CR_EXT; break;
	        case EX_QUIT: quit = TRUE; break;
	    }    
	} while( !quit );
	
	if( updated && db_support ) write_cddb( TRUE );
	done_display();
	close_cdrom();

	return( ERR_NOERR );
}